/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigator.h"

#include <cmath>
#include <iterator>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <Eigen/src/Geometry/Transform.h>

#include "ArmarXCore/core/services/tasks/TaskUtil.h"
#include "ArmarXCore/util/CPPUtility/trace.h"
#include "ArmarXGui/libraries/RemoteGui/Client/Widgets.h"

#include "RobotAPI/libraries/core/remoterobot/RemoteRobot.h"

// Navigation
#include "Navigation/libraries/core/types.h"
#include "Navigation/libraries/util/util.h"
#include <Navigation/libraries/factories/NavigationStackFactory.h>

armarx::nav::components::Navigator::Navigator() : executor{platformUnit}
{
}

armarx::nav::components::Navigator::~Navigator()
{
}

void armarx::nav::components::Navigator::onInitComponent()
{
}

void armarx::nav::components::Navigator::onConnectComponent()
{
    {
        createRemoteGuiTab();
        RemoteGui_startRunningTask();
    }
}

void armarx::nav::components::Navigator::onDisconnectComponent()
{
}

void armarx::nav::components::Navigator::onExitComponent()
{
}

void armarx::nav::components::Navigator::updateContext()
{
    scene.staticScene = staticScene();
    scene.robot = getRobot();
}

std::string armarx::nav::components::Navigator::getDefaultName() const
{
    return "Navigator";
}

std::vector<armarx::nav::core::Pose> convert(const std::vector<Eigen::Matrix4f>& wps)
{
    using namespace armarx::nav;
    std::vector<core::Pose> p;
    p.reserve(wps.size());
    std::transform(wps.begin(),
                   wps.end(),
                   std::back_inserter(p),
                   [](const auto & p)
    {
        return Eigen::Affine3f(p);
    });
    return p;
}

std::string
armarx::nav::components::Navigator::createConfig(const aron::data::AronDictPtr& stackConfig,
        const Ice::Current&)
{
    // TODO: Not thread-safe.
    const std::string uuid = boost::uuids::to_string(boost::uuids::random_generator()());

    server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, scene);
    navigators.emplace(uuid, server::Navigator{stack, scene, executor});

    return uuid;
}

void armarx::nav::components::Navigator::moveTo(const std::vector<Eigen::Matrix4f>& waypoints,
        const std::string& navigationMode,
        const std::string& configId,
        const Ice::Current&)
{
    // TODO: Error handling.

    navigators.at(configId).moveTo(convert(waypoints),
                                   core::NavigationFramesMap.from_name(navigationMode));
}

void armarx::nav::components::Navigator::moveTowards(const Eigen::Vector3f& direction,
        const std::string& navigationMode,
        const std::string& configId,
        const Ice::Current&)
{
    // TODO: Error handling.

    navigators.at(configId).moveTowards(direction,
                                        core::NavigationFramesMap.from_name(navigationMode));
}

void armarx::nav::components::Navigator::pauseMovement(const Ice::Current&)
{
}

void armarx::nav::components::Navigator::resumeMovement(const Ice::Current&)
{
}

armarx::PropertyDefinitionsPtr armarx::nav::components::Navigator::createPropertyDefinitions()
{
    ARMARX_TRACE;

    armarx::PropertyDefinitionsPtr def = new ComponentPropertyDefinitions(getConfigIdentifier());

    // Publish to a topic (passing the TopicListenerPrx).
    // def->topic(myTopicListener);

    // Subscribe to a topic (passing the topic name).
    // def->topic<PlatformUnitListener>("MyTopic");

    // Use (and depend on) another component (passing the ComponentInterfacePrx).
    // def->component(platformUnit);

    // Add a required property.
    // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

    // Add an optionalproperty.
    // def->optional(properties.numBoxes, "p.box.Number", "Number of boxes to draw in ArViz.");

    return def;
}

armarx::nav::core::StaticScene armarx::nav::components::Navigator::staticScene()
{
    ARMARX_TRACE;

    core::StaticScene scene;

    objpose::ObjectPoseSeq objectPoses = ObjectPoseClientPluginUser::getObjectPoses();

    scene.objects = util::asSceneObjects(objectPoses);

    return scene;
}

VirtualRobot::RobotPtr armarx::nav::components::Navigator::getRobot()
{
    ARMARX_TRACE;

    auto robot = RemoteRobot::createLocalCloneFromFile(
                     getRobotStateComponent(), VirtualRobot::RobotIO::RobotDescription::eFull);
    // auto robot = RemoteRobot::createLocalClone(getRobotStateComponent());
    ARMARX_CHECK_NOT_NULL(robot);

    // ARMARX_INFO << "Resetting robot";
    // robot = VirtualRobot::RobotIO::loadRobot(robot->getFilename());

    RemoteRobot::synchronizeLocalClone(robot, getRobotStateComponent());

    return robot;
}

void armarx::nav::components::Navigator::createRemoteGuiTab()
{
    ARMARX_TRACE;
    using namespace armarx::RemoteGui::Client;

    // Setup the widgets.

    {
        tab.targetPoseGroup.group.setLabel("Target pose");

        {
            float max = 10'000;
            tab.targetPoseGroup.targetPoseX.setRange(-max, max);
            tab.targetPoseGroup.targetPoseX.setDecimals(0);
            tab.targetPoseGroup.targetPoseX.setSteps(int(max / 10));
            tab.targetPoseGroup.targetPoseX.setValue(0.F);
        }

        {
            float max = 10'000;
            tab.targetPoseGroup.targetPoseY.setRange(-max, max);
            tab.targetPoseGroup.targetPoseY.setDecimals(0);
            tab.targetPoseGroup.targetPoseY.setSteps(int(max / 10));
            tab.targetPoseGroup.targetPoseY.setValue(0.F);
        }

        {
            float max = 10'000;
            tab.targetPoseGroup.targetPoseZ.setRange(-max, max);
            tab.targetPoseGroup.targetPoseZ.setDecimals(0);
            tab.targetPoseGroup.targetPoseZ.setSteps(int(max / 10));
            tab.targetPoseGroup.targetPoseZ.setValue(0.F);
        }

        {
            float max = 180;
            tab.targetPoseGroup.targetPoseRoll.setRange(-max, max);
            tab.targetPoseGroup.targetPoseRoll.setDecimals(0);
            tab.targetPoseGroup.targetPoseRoll.setSteps(2 * max);
            tab.targetPoseGroup.targetPoseRoll.setValue(0.F);
        }

        {
            float max = 180;
            tab.targetPoseGroup.targetPosePitch.setRange(-max, max);
            tab.targetPoseGroup.targetPosePitch.setDecimals(0);
            tab.targetPoseGroup.targetPosePitch.setSteps(2 * max);
            tab.targetPoseGroup.targetPosePitch.setValue(0.F);
        }

        {
            float max = 180;
            tab.targetPoseGroup.targetPoseYaw.setRange(-max, max);
            tab.targetPoseGroup.targetPoseYaw.setDecimals(0);
            tab.targetPoseGroup.targetPoseYaw.setSteps(2 * max);
            tab.targetPoseGroup.targetPoseYaw.setValue(0.F);
        }

        // tab.boxLayerName.setValue(0);
        // tab.numBoxes.setValue(1);
        // tab.numBoxes.setRange(0, 100);

        // tab.drawBoxes.setLabel("Draw Boxes");

        // Setup the layout.

        GridLayout grid;
        int row = 0;
        {

            tab.controlGroup.continueButton.setLabel("Continue");
            tab.controlGroup.pauseButton.setLabel("Pause");
            tab.controlGroup.moveToButton.setLabel("MoveTo");

            grid.add(Label("X"), {row, 0}).add(tab.targetPoseGroup.targetPoseX, {row, 1});
            ++row;

            grid.add(Label("Y"), {row, 0}).add(tab.targetPoseGroup.targetPoseY, {row, 1});
            ++row;

            grid.add(Label("Z"), {row, 0}).add(tab.targetPoseGroup.targetPoseZ, {row, 1});
            ++row;

            grid.add(Label("Roll"), {row, 0}).add(tab.targetPoseGroup.targetPoseRoll, {row, 1});
            ++row;

            grid.add(Label("Pitch"), {row, 0}).add(tab.targetPoseGroup.targetPosePitch, {row, 1});
            ++row;

            grid.add(Label("Yaw"), {row, 0}).add(tab.targetPoseGroup.targetPoseYaw, {row, 1});
            ++row;

            // grid.add(Label("Num Boxes"), {row, 0}).add(tab.numBoxes, {row, 1});
            // ++row;

            // grid.add(tab.drawBoxes, {row, 0}, {2, 1});
            // ++row;
        }

        VBoxLayout root = {grid, VSpacer()};

        tab.targetPoseGroup.group.addChild(root);
    }

    {
        tab.controlGroup.group.setLabel("Control");

        GridLayout grid;
        int row = 10;
        {
            grid.add(tab.controlGroup.moveToButton, {row, 1});
            ++row;

            grid.add(tab.controlGroup.pauseButton, {row, 1});
            ++row;

            grid.add(tab.controlGroup.continueButton, {row, 1});
            ++row;
        }

        VBoxLayout root = {grid, VSpacer()};
        tab.controlGroup.group.addChild(root);
    }

    ARMARX_TRACE;

    tab.hbox.addChild(tab.targetPoseGroup.group);
    tab.hbox.addChild(tab.controlGroup.group);

    RemoteGui_createTab(getName(), tab.hbox, &tab);
}

void armarx::nav::components::Navigator::RemoteGui_update()
{
    ARMARX_TRACE;

    // if (tab.boxLayerName.hasValueChanged() || tab.numBoxes.hasValueChanged())
    // {
    //     std::scoped_lock lock(propertiesMutex);
    //     properties.boxLayerName = tab.boxLayerName.getValue();
    //     properties.numBoxes = tab.numBoxes.getValue();

    //     {
    //         setDebugObserverDatafield("numBoxes", properties.numBoxes);
    //         setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
    //         sendDebugObserverBatch();
    //     }
    // }
    // if (tab.drawBoxes.wasClicked())
    // {
    //     ARMARX_INFO << "Navigator control requested from remote GUI";
    //     // Lock shared variables in methods running in seperate threads
    //     // and pass them to functions. This way, the called functions do
    //     // not need to think about locking.
    //     std::scoped_lock lock(propertiesMutex);
    //     // drawBoxes(properties, arviz);
    // }
}
